/* BEGIN NOTICE

This software was written by Peter DiFalco and John Sarraille.  The
algorithm employed was inspired by the paper entitled "A Fast
Algorithm To Determine Fractal Dimensions By Box Counting", which was
written by Liebovitch and Toth, and which appeared in the journal
"Physics Letters A", volume 141, pp 386-390, (1989).

This software is not warranteed: use at your own risk.  Unlimited
permission to distribute this software is granted, so long as this
notice is included in all copies made.

August 19, 1991

END NOTICE */

/*
fdqueue.c
   Contains queue package for Fractal Dimension Program:

empty_Q(Q) --	Tests to see if the queue Q is empty. Returns true if the queue
		rear points to NULL.
full_Q(Q) -- 	Tests to see if the queue Q is full. Returns false - a linked
		list is never full.
en_Q(Q,ptr) -- 	Places a new node containing the X_Y_marker "ptr" at the end
		of the queue Q. Q is a pointer to a queue, so no return is
		necessary.
de_Q(Q,ptr) -- 	Frees the front node from the queue Q and assigns its
		X_Y_marker to "ptr". "ptr" is passed by reference so no
		return is necessary.
create_Q(Q) -- 	Creates a queue Q and initializes the front and rear pointers
		to NULL.
destroy_Q(Q) -- Destroys the queue Q and frees all memory used by Q.

NOTE:  This package ought to be changed eventually, so that the same
memory used to load the data is used as the memory for the queues.
This would make the program use approximately half as much memory.
*/

#include <stdio.h>
#include "fd.h"

int empty_Q();
int full_Q();
void en_Q();
void de_Q();
void create_Q();
void destroy_Q();


int empty_Q(Q)
	QHeader Q; 	/* The queue to be tested. */
{
	return(Q.Qrear == NULL);
}

int full_Q(Q)
	QHeader Q;	/* The queue to be tested. */
{
	return(0);	/* A linked list is never full. */
}

void en_Q(Q,ptr)
	Queue Q;		/* Queue to enqueue. */
	X_Y_marker ptr;	/* Info to go into the node being enqueued. */
{
	QNode *newnode;	/* New node to be enqueued. */

   /* ############################## */
   /* When we run this on our local computer, the program fails somewhat
      mysteriously when the data set is over 400,000 points.  This error
      message is an effort to see if the problem is caused by running out
      of memory in the queues used for sorting the data.  */
   if ((newnode = (QNode *) malloc (sizeof(QNode))) == NULL)
   {
      printf("memory allocation failure -- en_Q function failing.\n");
      exit(-1);
   }

   /* This is the old form of the memory allocation command -- no error
      message in the case of failure.  */
   /* newnode = (QNode *) malloc (sizeof(QNode)); */
   
	newnode->next = NULL;
	(newnode->elt) = ptr;
	if (empty_Q(*Q))
		Q->Qfront = newnode;
	else
		Q->Qrear->next = newnode;
	Q->Qrear = newnode;
}

void de_Q(Q,ptr)
	Queue Q;		/* Queue to be dequeued. */
	X_Y_marker *ptr;	/* Pointer to the data to be returned */
				/* to the calling function. */	
{
	QNode *tempnode;	/* Temporary Queue node. */

	*ptr = (Q->Qfront->elt);
	tempnode = Q->Qfront;
	Q->Qfront = Q->Qfront->next;
	if (Q->Qfront == NULL)
		Q->Qrear = NULL;
	free(tempnode);
}

void create_Q(Q)
	Queue *Q;	/* Queue to be created. */
{
	*Q = (QHeader *) malloc (sizeof(QHeader));

	if (*Q == NULL)
		printf("ERROR: Out of space to create a queue. \n");
	else
	{
		(*Q)->Qrear = NULL;
		(*Q)->Qfront = NULL;
	}
}

void destroy_Q(Q)
	Queue Q;	/* Queue to be destroyed. */
{
	X_Y_marker ptr;	/* Pointer to data element of queue node. */

	while (!empty_Q(*Q))
		de_Q(Q,&ptr);

	/* destroy_Q seems to be working alright now that the line
	   below has been deleted.  The C compiler on the Sun
	   SparcStation did not seem to mind this line of code, but
	   the C compiler on the NeXT barfed. */

       /*free(ptr); */
}
